from __future__ import annotations

import uuid
from typing import Any, Type

import aiohttp

from butterflymx.graphql import Func, Q
from butterflymx.models import (
    AccessToken,
    EmailAndPassword,
    OauthCredentials,
    AccessPoint,
    Building,
    Tenant,
)
from butterflymx.request_client import ButterflyMXRequestClient


class ButterflyMX:
    def __init__(
            self,
            *,
            oauth_credentials: OauthCredentials | None = None,
            email_and_password: EmailAndPassword | None = None,
            access_token: AccessToken | None = None,
            refresh_token: str | None = None,
    ):
        self.__oauth_credentials = oauth_credentials
        self.__email_and_password = email_and_password
        self.__access_token = access_token
        self.__refresh_token = refresh_token

        self.__http: aiohttp.ClientSession | None = None
        self.__client: ButterflyMXRequestClient | None = None

    @property
    def client(self) -> ButterflyMXRequestClient:
        if self.__client is None:
            raise RuntimeError('You must use this class as an async context manager')

        return self.__client

    async def __aenter__(self) -> ButterflyMX:
        self.__http = aiohttp.ClientSession()

        self.__client = ButterflyMXRequestClient(
            http=self.__http,
            oauth_credentials=self.__oauth_credentials,
            email_and_password=self.__email_and_password,
            access_token=self.__access_token,
            refresh_token=self.__refresh_token,
        )

        return self

    async def __aexit__(
            self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: Any
    ) -> None:
        assert self.__http is not None
        await self.__http.close()
        self.__http = None

    async def graphql_query(self, query: str | Q) -> dict[str, Any]:
        response = await self.client.request(
            'POST',
            'https://api.butterflymx.com/denizen/v1/graphql',
            json={'query': str(query)},
        )

        data = await response.json()

        if len(data) == 1 and 'data' in data:
            data = data['data']

        return data

    async def get_types_introspection(self, *types: str) -> dict[str, Any]:
        """
        Retrieves information on the specified types from the ButterflyMX
        GraphQL API including their fields and information about those fields
        such as types and descriptions.
        """

        q_introspection = Q([
            Q([
                'name',
                Q(
                    ['name', Q(['name', 'kind', Q(['name', 'kind'], name='ofType')], name='type')],
                    name='fields',
                ),
            ], name=f'{t}: __type(name: "{t}")')
            for t in types
        ])

        return await self.graphql_query(q_introspection)

    async def get_tenants(self) -> list[Tenant]:
        """
        Retrieves a tenant (the logged-in user most likely) and related data
        """

        q_access_points = Q(
            Q(
                [
                    'id',
                    'legacyId',
                    'name',
                    'capabilities',
                    Q(['id', 'name'], name='building'),
                ],
                name='nodes'
            ),
            name='accessPoints',
        )

        q_tenants_access_points = Q(Q(
            Q(['id', 'name', q_access_points], name='nodes'),
            name='tenants',
        ))

        data = await self.graphql_query(q_tenants_access_points)

        return [
            Tenant(
                id=tenant['id'],
                name=tenant['name'],
                access_points=[
                    AccessPoint(
                        id=access_point['id'],
                        legacy_id=access_point['legacyId'],
                        name=access_point['name'],
                        building=Building(
                            id=access_point['building']['id'],
                            name=access_point['building']['name'],
                        ),
                        capabilities=access_point['capabilities'],
                    )
                    for access_point in tenant['accessPoints']['nodes']
                ],
            )
            for tenant in data['tenants']['nodes']
        ]

    async def open_access_point(
            self,
            tenant: str | Tenant,
            access_point: str | AccessPoint
    ) -> None:
        tenant = tenant if isinstance(tenant, str) else tenant.id
        access_point = access_point if isinstance(access_point, str) else access_point.id

        await self.graphql_query(
            query=Q(
                Func(
                    "swipeToOpen",
                    {
                        "input": {
                            "tenantId": tenant,
                            "accessPointId": access_point,
                            "deviceId": None,
                            "clientMutationId": str(uuid.uuid4()),
                        }
                    },
                    ['clientMutationId'],
                ),
                name="mutation",
            )
        )



from urllib.parse import parse_qs, urlparse


def get_query_params(url: str) -> dict[str, str | list[str]]:
    return {k: (v[0] if len(v) == 1 else v) for k, v in parse_qs(urlparse(url).query).items()}



from __future__ import annotations

import abc
import json
from typing import Any, Sequence

DEFAULT_INDENTATION = 2


class AbstractBuilder(abc.ABC):
    @abc.abstractmethod
    def render(self, *, indent: int = DEFAULT_INDENTATION, depth: int = 1) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError


def render_fields(
    *,
    fields: Sequence[str | AbstractBuilder],
    depth: int,
    indent: int,
) -> str:
    indent_str = ' ' * (depth * indent)

    return '\n'.join(
        indent_str + (f.render(depth=(depth + 1)) if isinstance(f, AbstractBuilder) else f)
        for f in fields
    )


def render_dict(
    *,
    fields: dict[str, str | dict[str, Any]],
    depth: int,
    indent: int,
    root: bool = False,
) -> str:
    depth -= root

    indent_str = ' ' * (depth * indent)

    def render_field(value: Any) -> str:
        return render_dict(
            fields=value,
            depth=(depth + 1),
            indent=indent
        ) if isinstance(value, dict) else json.dumps(value)

    rendered = render_fields(
        fields=[
            f'{k}: {render_field(v)}'
            for k, v in fields.items()
        ],
        depth=depth,
        indent=indent,
    )

    if root:
        return rendered

    return f'{{\n{rendered}\n{indent_str[:indent * (depth - 1)]}}}'


class Q(AbstractBuilder):
    def __init__(
        self,
        fields: Sequence[str | AbstractBuilder] | str | AbstractBuilder | None = None,
        name: str = 'query',
    ):
        self.fields: Sequence[str | AbstractBuilder] = (
            [fields] if isinstance(fields, (str, AbstractBuilder)) else (fields or [])
        )
        self.name = name

    def render(self, *, depth: int = 1, indent: int = DEFAULT_INDENTATION) -> str:
        indent_str = ' ' * (depth * indent)

        fields = render_fields(fields=self.fields, depth=depth, indent=indent)

        return f'{self.name} {{\n{fields}\n{indent_str[:(depth - 1) * indent]}}}'

    def __str__(self) -> str:
        return self.render()


class Func(AbstractBuilder):
    def __init__(
        self,
        name: str,
        arguments: dict[str, Any] | None = None,
        selections: Sequence[str | Q] | None = None,
    ):
        self.name = name
        self.arguments = arguments or {}
        self.selections = selections or []

    def render(self, *, depth: int = 1, indent: int = DEFAULT_INDENTATION) -> str:
        indent_str = ' ' * (depth * indent)

        rendered_arguments = render_dict(
            fields=self.arguments,
            depth=depth,
            indent=indent,
            root=True,
        ).lstrip()

        rendered_selections = render_fields(fields=self.selections, depth=depth, indent=indent)

        return (
            f'{self.name}({rendered_arguments}) {{\n{rendered_selections}\n'
            f'{indent_str[:(depth - 1) * indent]}}}'
        )

    def __str__(self) -> str:
        return self.render()



from dataclasses import dataclass


@dataclass(frozen=True, kw_only=True, slots=True)
class EmailAndPassword:
    email: str
    password: str


@dataclass(frozen=True, kw_only=True, slots=True)
class OauthCredentials:
    """
    Oauth credentials used for login flow, these can be obtained by using a debugging proxy while
    logging in via the mobile app.
    """

    client_id: str
    client_secret: str


@dataclass(frozen=True, kw_only=True, slots=True)
class AccessToken:
    token: str
    expires_at: float



from dataclasses import dataclass


@dataclass(frozen=True, kw_only=True, slots=True)
class Building:
    id: str
    name: str


@dataclass(frozen=True, kw_only=True, slots=True)
class AccessPoint:
    id: str
    legacy_id: str
    name: str
    building: Building
    capabilities: list[str]


@dataclass(frozen=True, kw_only=True, slots=True)
class Tenant:
    id: str
    name: str
    access_points: list[AccessPoint]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['butterflymx', 'butterflymx.graphql', 'butterflymx.models']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.4,<4.0.0',
 'beautifulsoup4>=4.12.2,<5.0.0',
 'ruff>=0.0.265,<0.0.266']

setup_kwargs = {
    'name': 'butterflymx',
    'version': '0.0.2',
    'description': 'A reverse-engineered ButterflyMX app API wrapper',
    'long_description': '# ButterflyMX\nA reverse-engineered ButterflyMX app API wrapper\n',
    'author': 'Milo Weinberg',
    'author_email': 'iapetus011@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import time
from typing import Any

import aiohttp
from aiohttp import FormData
import bs4

from butterflymx.models import EmailAndPassword
from butterflymx.models.login import AccessToken, OauthCredentials
from butterflymx.utils import get_query_params

SPOOF_USER_AGENT = (
    'Mozilla/5.0 (iPhone; CPU iPhone OS 16_1_1 like Mac OS X) '
    'AppleWebKit/605.1.15 (KHTML, like Gecko) Mobile/15E148'
)


class ButterflyMXRequestClient:
    def __init__(
        self,
        *,
        http: aiohttp.ClientSession,
        oauth_credentials: OauthCredentials | None = None,
        email_and_password: EmailAndPassword | None = None,
        access_token: AccessToken | None = None,
        refresh_token: str | None = None,
    ):
        if not (refresh_token or email_and_password):
            raise TypeError('Either refresh_token or email_and_password must be specified.')

        self.__http = http
        self.__oauth_credentials = oauth_credentials
        self.__email_and_password = email_and_password
        self.__refresh_token = refresh_token

        self.__access_token: str | None = None
        self.__access_token_expires_at: float | None = None

        if access_token is not None:
            self.__access_token = access_token.token
            self.__access_token_expires_at = access_token.expires_at

    @property
    def access_token(self) -> str:
        if self.__access_token is None:
            raise RuntimeError('access_token has not yet been initialized.')

        return self.__access_token

    @property
    def access_token_expires_at(self) -> float:
        if self.__access_token_expires_at is None:
            raise RuntimeError('access_token_expires_at has not yet been initialized.')

        return self.__access_token_expires_at

    @property
    def refresh_token(self) -> str:
        if self.__refresh_token is None:
            raise RuntimeError('refresh_token has not yet been initialized.')

        return self.__refresh_token

    async def __oauth_login(self) -> tuple[AccessToken, str]:
        if not self.__email_and_password:
            raise RuntimeError('Oauth login requires email and password to be supplied.')

        if not self.__oauth_credentials:
            raise RuntimeError('Oauth login requires oauth credentials to be supplied.')

        # Start oauth flow
        await self.request(
            'GET',
            'https://accounts.butterflymx.com/oauth/authorize',
            params={
                'client_id': self.__oauth_credentials.client_id,
                'redirect_uri': 'com.butterflymx.oauth://oauth',
                'response_type': 'code',
            },
            authenticate=False,
            allow_redirects=False,
        )

        # Fetch authenticity token from login page
        # 3
        response = await self.request(
            'GET', 'https://accounts.butterflymx.com/login/new', authenticate=False
        )

        authenticity_token_input = bs4.BeautifulSoup(await response.text(), 'html.parser').find(
            'input', attrs={'name': 'authenticity_token'}
        )
        assert isinstance(authenticity_token_input, bs4.Tag)

        # Submit email and password
        # 7
        response = await self.request(
            'POST',
            'https://accounts.butterflymx.com/login',
            data=FormData(
                {
                    'utf8': '✓',  # Wtf is this
                    'authenticity_token': authenticity_token_input.attrs['value'],
                    'account[email]': self.__email_and_password.email,
                    'account[password]': self.__email_and_password.password,
                    'button': '',
                }
            ),
            headers={
                'Referer': 'https://accounts.butterflymx.com/login/new',
                'Origin': 'https://accounts.butterflymx.com',
                'User-Agent': SPOOF_USER_AGENT,
            },
            authenticate=False,
            allow_redirects=False,
        )

        # Oauth flow
        # 8
        response = await self.request(
            'GET',
            response.headers['Location'],
            headers={
                'User-Agent': SPOOF_USER_AGENT,
                'Referer': 'https://accounts.butterflymx.com/login/new',
            },
            authenticate=False,
            allow_redirects=False,
        )
        oauth_code = get_query_params(response.headers['Location'])['code']

        # Retrieve access and refresh tokens
        response = await self.request(
            'POST',
            'https://accounts.butterflymx.com/oauth/token',
            params={
                'client_secret': self.__oauth_credentials.client_secret,
                'grant_type': 'authorization_code',
                'client_id': self.__oauth_credentials.client_id,
                'redirect_uri': 'com.butterflymx.oauth://oauth',
                'code': oauth_code,
            },
            authenticate=False,
            allow_redirects=False,
        )
        token_data = await response.json()

        return (
            AccessToken(
                token=token_data['access_token'],
                expires_at=(token_data['created_at'] + token_data['expires_in']),
            ),
            token_data['refresh_token'],
        )

    async def __oauth_refresh(self) -> tuple[AccessToken, str]:
        if not self.__oauth_credentials:
            raise RuntimeError('Oauth login requires oauth credentials to be supplied.')

        response = await self.request(
            'POST',
            'https://accounts.butterflymx.com/oauth/token',
            json={
                'refresh_token': self.__refresh_token,
                'client_id': self.__oauth_credentials.client_id,
                'grant_type': 'refresh_token',
            },
            authenticate=False,
        )
        token_data = await response.json()

        return (
            AccessToken(
                token=token_data['access_token'],
                expires_at=(token_data['created_at'] + token_data['expires_in']),
            ),
            token_data['refresh_token'],
        )

    async def ensure_access_token(self) -> None:
        if (
            self.__access_token is None
            or self.__access_token_expires_at is None
            or time.time() > self.__access_token_expires_at
        ):
            access_token: AccessToken
            refresh_token: str

            if self.__refresh_token is None:
                (access_token, refresh_token) = await self.__oauth_login()
            else:
                (access_token, refresh_token) = await self.__oauth_refresh()

            self.__access_token = access_token.token
            self.__access_token_expires_at = access_token.expires_at
            self.__refresh_token = refresh_token

    async def request(
        self,
        method: str,
        url: str,
        *,
        json: Any = None,
        data: Any = None,
        params: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        authenticate: bool = True,
        **kwargs,
    ) -> aiohttp.ClientResponse:
        if headers is None:
            headers = {}

        if authenticate:
            await self.ensure_access_token()
            headers['Authorization'] = self.access_token

        return await self.__http.request(
            method, url, data=data, json=json, params=params, headers=headers, **kwargs
        )
